@@ -37,25 +37,17 @@
 // of a compare-and-swap which is expensive).
 
 // SpinLock is async signal safe.
-// If used within a signal handler, all lock holders 
+// If used within a signal handler, all lock holders
 // should block the signal even outside the signal handler.
 
 #ifndef BASE_SPINLOCK_H_
 #define BASE_SPINLOCK_H_
 
-#include "config.h"
+#include <config.h>
 #include "base/basictypes.h"
 #include "base/atomicops.h"
 #include "base/dynamic_annotations.h"
-
-// One day, we may use __attribute__ stuff on gcc to annotate these functions
-#define LOCKABLE
-#define SCOPED_LOCKABLE
-#define EXCLUSIVE_LOCK_FUNCTION(...)
-#define EXCLUSIVE_TRYLOCK_FUNCTION(...)
-#define UNLOCK_FUNCTION(...)
-
-
+#include "base/thread_annotations.h"
 
 class LOCKABLE SpinLock {
  public:
@@ -70,12 +62,14 @@ class LOCKABLE SpinLock {
   // A SpinLock constructed like this can be freely used from global
   // initializers without worrying about the order in which global
   // initializers run.
-  explicit SpinLock(base::LinkerInitialized x) {
+  explicit SpinLock(base::LinkerInitialized /*x*/) {
     // Does nothing; lockword_ is already initialized
   }
 
   // Acquire this SpinLock.
-  inline void Lock() EXCLUSIVE_LOCK_FUNCTION() {
+  // TODO(csilvers): uncomment the annotation when we figure out how to
+  //                 support this macro with 0 args (see thread_annotations.h)
+  inline void Lock() /*EXCLUSIVE_LOCK_FUNCTION()*/ {
     if (Acquire_CompareAndSwap(&lockword_, 0, 1) != 0) {
       SlowLock();
     }
@@ -95,7 +89,9 @@ class LOCKABLE SpinLock {
   }
 
   // Release this SpinLock, which must be held by the calling thread.
-  inline void Unlock() UNLOCK_FUNCTION() {
+  // TODO(csilvers): uncomment the annotation when we figure out how to
+  //                 support this macro with 0 args (see thread_annotations.h)
+  inline void Unlock() /*UNLOCK_FUNCTION()*/ {
     // This is defined in mutex.cc.
     extern void SubmitSpinLockProfileData(const void *, int64);
 
@@ -152,7 +148,9 @@ class SCOPED_LOCKABLE SpinLockHolder {
       : lock_(l) {
     l->Lock();
   }
-  inline ~SpinLockHolder() UNLOCK_FUNCTION() { lock_->Unlock(); }
+  // TODO(csilvers): uncomment the annotation when we figure out how to
+  //                 support this macro with 0 args (see thread_annotations.h)
+  inline ~SpinLockHolder() /*UNLOCK_FUNCTION()*/ { lock_->Unlock(); }
 };
 // Catch bug where variable name is omitted, e.g. SpinLockHolder (&lock);
 #define SpinLockHolder(x) COMPILE_ASSERT(0, spin_lock_decl_missing_var_name)
