@@ -36,10 +36,10 @@
 #ifndef _SYSINFO_H_
 #define _SYSINFO_H_
 
-#include "config.h"
+#include <config.h>
 
 #include <time.h>
-#ifdef WIN32
+#if (defined(_WIN32) || defined(__MINGW32__)) && (!defined(__CYGWIN__) && !defined(__CYGWIN32__))
 #include <windows.h>   // for DWORD
 #include <TlHelp32.h>  // for CreateToolhelp32Snapshot
 #endif
@@ -49,15 +49,21 @@
 #include <stddef.h>    // for size_t
 #include <limits.h>    // for PATH_MAX
 #include "base/basictypes.h"
+#include "base/logging.h"   // for RawFD
 
-// This getenv prefers using /proc/self/environ to calling getenv().
-// It's intended to be used in routines that run before main(), when
-// the state required for getenv() may not be set up yet.  In particular,
-// errno isn't set up until relatively late (after the pthreads library
-// has a chance to make it threadsafe), and getenv() doesn't work until then.
-// Note that /proc only has the environment at the time the application was
-// started, so this routine ignores setenv() calls/etc.  Also note it only
-// reads the first 16K of the environment.
+// This getenv function is safe to call before the C runtime is initialized.
+// On Windows, it utilizes GetEnvironmentVariable() and on unix it uses
+// /proc/self/environ instead calling getenv().  It's intended to be used in
+// routines that run before main(), when the state required for getenv() may
+// not be set up yet.  In particular, errno isn't set up until relatively late
+// (after the pthreads library has a chance to make it threadsafe), and
+// getenv() doesn't work until then. 
+// On some platforms, this call will utilize the same, static buffer for
+// repeated GetenvBeforeMain() calls. Callers should not expect pointers from
+// this routine to be long lived.
+// Note that on unix, /proc only has the environment at the time the
+// application was started, so this routine ignores setenv() calls/etc.  Also
+// note it only reads the first 16K of the environment.
 extern const char* GetenvBeforeMain(const char* name);
 
 // This takes as an argument an environment-variable name (like
@@ -72,6 +78,14 @@ extern int NumCPUs();
 extern double CyclesPerSecond(void);
 
 
+//  Return true if we're running POSIX (e.g., NPTL on Linux) threads,
+//  as opposed to a non-POSIX thread libary.  The thing that we care
+//  about is whether a thread's pid is the same as the thread that
+//  spawned it.  If so, this function returns true.
+//  Thread-safe.
+//  Note: We consider false negatives to be OK.
+bool HasPosixThreads();
+
 #ifndef SWIG  // SWIG doesn't like struct Buffer and variable arguments.
 
 // A ProcMapsIterator abstracts access to /proc/maps for a given
@@ -181,7 +195,7 @@ class ProcMapsIterator {
   char *etext_;       // end of text
   char *nextline_;    // start of next line
   char *ebuf_;        // end of buffer (1 char for a nul)
-#if defined(WIN32)
+#if (defined(_WIN32) || defined(__MINGW32__)) && (!defined(__CYGWIN__) && !defined(__CYGWIN32__))
   HANDLE snapshot_;   // filehandle on dll info
   // In a change from the usual W-A pattern, there is no A variant of
   // MODULEENTRY32.  Tlhelp32.h #defines the W variant, but not the A.
@@ -208,4 +222,11 @@ class ProcMapsIterator {
 
 #endif  /* #ifndef SWIG */
 
+// Helper routines
+
+namespace tcmalloc {
+int FillProcSelfMaps(char buf[], int size);
+void DumpProcSelfMaps(RawFD fd);
+}
+
 #endif   /* #ifndef _SYSINFO_H_ */
