@@ -58,7 +58,7 @@
 
 // Annoying stuff for windows -- makes sure clients can import these functions
 #ifndef PERFTOOLS_DLL_DECL
-# ifdef WIN32
+# ifdef _WIN32
 #   define PERFTOOLS_DLL_DECL  __declspec(dllimport)
 # else
 #   define PERFTOOLS_DLL_DECL
@@ -120,41 +120,22 @@ class PERFTOOLS_DLL_DECL HeapLeakChecker {
   // has been called at least once).
   ~HeapLeakChecker();
 
-  // Return true iff the heap does not have more objects allocated
-  // w.r.t. its state at the time of our construction.
-  // This does full pprof heap change checking and reporting.
-  // To detect tricky leaks it depends on correct working pprof implementation
-  // referred by FLAGS_heap_profile_pprof.
-  // (By 'tricky leaks' we mean a change of heap state that e.g. for SameHeap
-  //  preserves the number of allocated objects and bytes
-  //  -- see TestHeapLeakCheckerTrick in heap-checker_unittest.cc --
-  //  and thus is not detected by BriefNoLeaks.)
-  // CAVEAT: pprof will do no checking over stripped binaries
-  // (our automatic test binaries are stripped)
-  // NOTE: All *NoLeaks() and *SameHeap() methods can be called many times
-  // to check for leaks at different end-points in program's execution.
-  bool NoLeaks() { return DoNoLeaks(NO_LEAKS, USE_PPROF, PPROF_REPORT); }
+  // These used to be different but are all the same now: they return
+  // true iff all memory allocated since this HeapLeakChecker object
+  // was constructor is still reachable from global state.
+  //
+  // Because we fork to convert addresses to symbol-names, and forking
+  // is not thread-safe, and we may be called in a threaded context,
+  // we do not try to symbolize addresses when called manually.
+  bool NoLeaks() { return DoNoLeaks(DO_NOT_SYMBOLIZE); }
 
-  // Return true iff the heap does not seem to have more objects allocated
-  // w.r.t. its state at the time of our construction
-  // by looking at the number of objects & bytes allocated.
-  // This also tries to do pprof reporting of detected leaks.
-  bool QuickNoLeaks() { return DoNoLeaks(NO_LEAKS, USE_COUNTS, PPROF_REPORT); }
-
-  // Return true iff the heap does not seem to have more objects allocated
-  // w.r.t. its state at the time of our construction
-  // by looking at the number of objects & bytes allocated.
-  // This does not try to use pprof at all.
-  bool BriefNoLeaks() { return DoNoLeaks(NO_LEAKS, USE_COUNTS, NO_REPORT); }
-
-  // These are similar to their *NoLeaks counterparts,
-  // but they in addition require no negative leaks,
-  // i.e. the state of the heap must be exactly the same
-  // as at the time of our construction.
-  bool SameHeap() { return DoNoLeaks(SAME_HEAP, USE_PPROF, PPROF_REPORT); }
-  bool QuickSameHeap()
-    { return DoNoLeaks(SAME_HEAP, USE_COUNTS, PPROF_REPORT); }
-  bool BriefSameHeap() { return DoNoLeaks(SAME_HEAP, USE_COUNTS, NO_REPORT); }
+  // These forms are obsolete; use NoLeaks() instead.
+  // TODO(csilvers): mark with ATTRIBUTE_DEPRECATED.
+  bool QuickNoLeaks()  { return NoLeaks(); }
+  bool BriefNoLeaks()  { return NoLeaks(); }
+  bool SameHeap()      { return NoLeaks(); }
+  bool QuickSameHeap() { return NoLeaks(); }
+  bool BriefSameHeap() { return NoLeaks(); }
 
   // Detailed information about the number of leaked bytes and objects
   // (both of these can be negative as well).
@@ -170,91 +151,28 @@ class PERFTOOLS_DLL_DECL HeapLeakChecker {
   // ----------------------------------------------------------------------- //
   // Static helpers to make us ignore certain leaks.
 
-  // NOTE: All calls to DisableChecks* affect all later heap profile generation
-  // that happens in our construction and inside of *NoLeaks().
-  // They do nothing when heap leak checking is turned off.
-
-  // CAVEAT: Disabling via all the DisableChecks* functions happens only
-  // up to kMaxStackTrace stack frames (see heap-profile-table.h)
-  // down from the stack frame identified by the function.
-  // Hence, this disabling will stop working for very deep call stacks
-  // and you might see quite wierd leak profile dumps in such cases.
-
-  // CAVEAT: Disabling via DisableChecksIn works only with non-strip'ped
-  // binaries.  It's better not to use this function if at all possible.
+  // Scoped helper class.  Should be allocated on the stack inside a
+  // block of code.  Any heap allocations done in the code block
+  // covered by the scoped object (including in nested function calls
+  // done by the code block) will not be reported as leaks.  This is
+  // the recommended replacement for the GetDisableChecksStart() and
+  // DisableChecksToHereFrom() routines below.
   //
-  // Register 'pattern' as another variant of a regular expression to match
-  // function_name, file_name:line_number, or function_address
-  // of function call/return points for which allocations below them should be
-  // ignored during heap leak checking.
-  // (This becomes a part of pprof's '--ignore' argument.)
-  // Usually this should be caled from a REGISTER_HEAPCHECK_CLEANUP
-  // in the source file that is causing the leaks being ignored.
-  static void DisableChecksIn(const char* pattern);
-
-  // A pair of functions to disable heap checking between them.
-  // For example
-  //    ...
-  //    void* start_address = HeapLeakChecker::GetDisableChecksStart();
-  //    <do things>
-  //    HeapLeakChecker::DisableChecksToHereFrom(start_address);
-  //    ...
-  // will disable heap leak checking for everything that happens
-  // during any execution of <do things> (including any calls from it),
-  // i.e. all objects allocated from there
-  // and everything reachable from them will not be considered a leak.
-  // Each such pair of function calls must be from the same function,
-  // because this disabling works by remembering the range of
-  // return program counter addresses for the two calls.
-  static void* GetDisableChecksStart();
-  static void DisableChecksToHereFrom(const void* start_address);
-
-  // ADVICE: Use GetDisableChecksStart, DisableChecksToHereFrom
-  //         instead of DisableChecksUp|At whenever possible
-  //         to make the code less fragile under different degrees of inlining.
-  // Register the function call point (return program counter address)
-  // 'stack_frames' above us for which allocations
-  // (everything reachable from them) below it should be
-  // ignored during heap leak checking.
-  // 'stack_frames' must be >= 1 (in most cases one would use the value of 1).
-  // For example
-  //    void Foo() {  // Foo() should not get inlined
-  //      HeapLeakChecker::DisableChecksUp(1);
-  //      <do things>
-  //    }
-  // will disable heap leak checking for everything that happens
-  // during any execution of <do things> (including any calls from it),
-  // i.e. all objects allocated from there
-  // and everything reachable from them will not be considered a leak.
-  // CAVEAT: If Foo() is inlined this will disable heap leak checking
-  // under all processing of all functions Foo() is inlined into.
-  // Hence, for potentially inlined functions, use the GetDisableChecksStart,
-  // DisableChecksToHereFrom calls instead.
-  // (In the above example we store and use the return program counter
-  //  addresses from Foo to do the disabling.)
-  static void DisableChecksUp(int stack_frames);
-
-  // Same as DisableChecksUp,
-  // but the function return program counter address is given explicitly.
-  static void DisableChecksAt(const void* address);
-
-  // Tests for checking that DisableChecksUp and DisableChecksAt
-  // behaved as expected, for example
-  //    void Foo() {
-  //      HeapLeakChecker::DisableChecksUp(1);
-  //      <do things>
-  //    }
-  //    void Bar() {
-  //      Foo();
-  //      CHECK(!HeapLeakChecker::HaveDisabledChecksUp(1));
-  //        // This will fail if Foo() got inlined into Bar()
-  //        // (due to more aggressive optimization in the (new) compiler)
-  //        // which breaks the intended behavior of DisableChecksUp(1) in it.
-  //      <do things>
-  //    }
-  // These return false when heap leak checking is turned off.
-  static bool HaveDisabledChecksUp(int stack_frames);
-  static bool HaveDisabledChecksAt(const void* address);
+  // Example:
+  //   void Foo() {
+  //     HeapLeakChecker::Disabler disabler;
+  //     ... code that allocates objects whose leaks should be ignored ...
+  //   }
+  //
+  // REQUIRES: Destructor runs in same thread as constructor
+  class Disabler {
+   public:
+    Disabler();
+    ~Disabler();
+   private:
+    Disabler(const Disabler&);        // disallow copy
+    void operator=(const Disabler&);  // and assign
+  };
 
   // Ignore an object located at 'ptr' (can go at the start or into the object)
   // as well as all heap objects (transitively) referenced from it
@@ -287,38 +205,37 @@ class PERFTOOLS_DLL_DECL HeapLeakChecker {
   // ----------------------------------------------------------------------- //
   // Various helpers
 
-  // Type for DumpProfileLocked
-  enum ProfileType { START_PROFILE, END_PROFILE };
+  // Create the name of the heap profile file.
+  // Should be deleted via Allocator::Free().
+  char* MakeProfileNameLocked();
 
-  // Helper for dumping start/end heap leak checking profiles
-  // and getting the byte/object counts.
-  void DumpProfileLocked(ProfileType profile_type, const void* self_stack_top,
-                         size_t* alloc_bytes, size_t* alloc_objects);
   // Helper for constructors
-  void Create(const char *name);
+  void Create(const char *name, bool make_start_snapshot);
 
-  // Types for DoNoLeaks and its helpers.
-  enum CheckType { SAME_HEAP, NO_LEAKS };
-  enum CheckFullness { USE_PPROF, USE_COUNTS };
-  enum ReportMode { PPROF_REPORT, NO_REPORT };
+  enum ShouldSymbolize { SYMBOLIZE, DO_NOT_SYMBOLIZE };
 
-  // Helpers for *NoLeaks and *SameHeap
-  bool DoNoLeaks(CheckType check_type,
-                 CheckFullness fullness,
-                 ReportMode report_mode);
-  struct LeakCheckResult;  // defined in .cc
-  LeakCheckResult DoLeakCheckLocked();
-  bool DoNoLeaksOnceLocked(CheckType check_type,
-                           CheckFullness fullness,
-                           ReportMode report_mode);
+  // Helper for *NoLeaks and *SameHeap
+  bool DoNoLeaks(ShouldSymbolize should_symbolize);
+
+  // These used to be public, but they are now deprecated.
+  // Will remove entirely when all internal uses are fixed.
+  // In the meantime, use friendship so the unittest can still test them.
+  static void* GetDisableChecksStart();
+  static void DisableChecksToHereFrom(const void* start_address);
+  static void DisableChecksIn(const char* pattern);
+  friend void RangeDisabledLeaks();
+  friend void NamedTwoDisabledLeaks();
+  friend void* RunNamedDisabledLeaks(void*);
+  friend void TestHeapLeakCheckerNamedDisabling();
+  friend int main(int, char**);
 
-  // Helper for DisableChecksAt
-  static void DisableChecksAtLocked(const void* address);
 
   // Helper for DisableChecksIn
   static void DisableChecksInLocked(const char* pattern);
 
-  // Helper for DisableChecksToHereFrom
+  // Disable checks based on stack trace entry at a depth <=
+  // max_depth.  Used to hide allocations done inside some special
+  // libraries.
   static void DisableChecksFromToLocked(const void* start_address,
                                         const void* end_address,
                                         int max_depth);
@@ -427,12 +344,17 @@ class PERFTOOLS_DLL_DECL HeapLeakChecker {
   class SpinLock* lock_;  // to make HeapLeakChecker objects thread-safe
   const char* name_;  // our remembered name (we own it)
                       // NULL means this leak checker is a noop
-  size_t start_inuse_bytes_;  // bytes in use at our construction
-  size_t start_inuse_allocs_;  // allocations in use at our construction
+
+  // Snapshot taken when the checker was created.  May be NULL
+  // for the global heap checker object.  We use void* instead of
+  // HeapProfileTable::Snapshot* to avoid including heap-profile-table.h.
+  void* start_snapshot_;
+
   bool has_checked_;  // if we have done the leak check, so these are ready:
   ssize_t inuse_bytes_increase_;  // bytes-in-use increase for this checker
   ssize_t inuse_allocs_increase_;  // allocations-in-use increase
                                    // for this checker
+  bool keep_profiles_;  // iff we should keep the heap profiles we've made
 
   // ----------------------------------------------------------------------- //
 
