@@ -37,6 +37,7 @@
 
 #include "addressmap-inl.h"
 #include "base/basictypes.h"
+#include "base/logging.h"   // for RawFD
 
 // Table to maintain a heap profile data inside,
 // i.e. the set of currently active heap memory allocations.
@@ -51,8 +52,8 @@ class HeapProfileTable {
   // Extension to be used for heap pforile files.
   static const char kFileExt[];
 
-  // Longest stack trace we record.
-  static const int kMaxStackDepth = 32;
+  // Longest stack trace we record.  Defined in the .cc file.
+  static const int kMaxStackDepth;
 
   // data types ----------------------------
 
@@ -75,6 +76,16 @@ class HeapProfileTable {
     size_t object_size;  // size of the allocation
     const void* const* call_stack;  // call stack that made the allocation call
     int stack_depth;  // depth of call_stack
+    bool live;
+    bool ignored;
+  };
+
+  // Info we return about an allocation context.
+  // An allocation context is a unique caller stack trace
+  // of an allocation operation.
+  struct AllocContextInfo : public Stats {
+    int stack_depth;                // Depth of stack trace
+    const void* const* call_stack;  // Stack trace
   };
 
   // Memory (de)allocator interface we'll use.
@@ -114,10 +125,14 @@ class HeapProfileTable {
 
   // If "ptr" points to a recorded allocation and it's not marked as live
   // mark it as live and return true. Else return false.
-  // All allocations start as non-live, DumpNonLiveProfile below
-  // also makes all allocations non-live.
+  // All allocations start as non-live.
   bool MarkAsLive(const void* ptr);
 
+  // If "ptr" points to a recorded allocation, mark it as "ignored".
+  // Ignored objects are treated like other objects, except that they
+  // are skipped in heap checking reports.
+  void MarkAsIgnored(const void* ptr);
+
   // Return current total (de)allocation statistics.
   const Stats& total() const { return total_; }
 
@@ -131,6 +146,14 @@ class HeapProfileTable {
     allocation_->Iterate(MapArgsAllocIterator, callback);
   }
 
+  // Allocation context profile data iteration callback
+  typedef void (*AllocContextIterator)(const AllocContextInfo& info);
+
+  // Iterate over the allocation context profile data calling "callback"
+  // for every allocation context. Allocation contexts are ordered by the
+  // size of allocated space.
+  void IterateOrderedAllocContexts(AllocContextIterator callback) const;
+
   // Fill profile data into buffer 'buf' of size 'size'
   // and return the actual size occupied by the dump in 'buf'.
   // The profile buckets are dumped in the decreasing order
@@ -138,24 +161,26 @@ class HeapProfileTable {
   // We do not provision for 0-terminating 'buf'.
   int FillOrderedProfile(char buf[], int size) const;
 
-  // Allocation data dump filtering callback:
-  // gets passed object pointer and size
-  // needs to return true iff the object is to be filtered out of the dump.
-  typedef bool (*DumpFilter)(const void* ptr, size_t size);
-
-  // Dump non-live portion of the current heap profile
-  // for leak checking purposes to file_name.
-  // Also reset all objects to non-live state
-  // and write the sums of allocated byte and object counts in the dump
-  // to *alloc_bytes and *alloc_objects.
-  // dump_alloc_addresses controls if object addresses are dumped.
-  bool DumpNonLiveProfile(const char* file_name,
-                          bool dump_alloc_addresses,
-                          Stats* profile_stats) const;
-
   // Cleanup any old profile files matching prefix + ".*" + kFileExt.
   static void CleanupOldProfiles(const char* prefix);
 
+  // Return a snapshot of the current contents of *this.
+  // Caller must call ReleaseSnapshot() on result when no longer needed.
+  // The result is only valid while this exists and until
+  // the snapshot is discarded by calling ReleaseSnapshot().
+  class Snapshot;
+  Snapshot* TakeSnapshot();
+
+  // Release a previously taken snapshot.  snapshot must not
+  // be used after this call.
+  void ReleaseSnapshot(Snapshot* snapshot);
+
+  // Return a snapshot of every non-live, non-ignored object in *this.
+  // If "base" is non-NULL, skip any objects present in "base".
+  // As a side-effect, clears the "live" bit on every live object in *this.
+  // Caller must call ReleaseSnapshot() on result when no longer needed.
+  Snapshot* NonLiveSnapshot(Snapshot* base);
+
  private:
 
   // data types ----------------------------
@@ -173,17 +198,32 @@ class HeapProfileTable {
   struct AllocValue {
     // Access to the stack-trace bucket
     Bucket* bucket() const {
-      return reinterpret_cast<Bucket*>(bucket_rep & ~uintptr_t(1));
+      return reinterpret_cast<Bucket*>(bucket_rep & ~uintptr_t(kMask));
     }
     // This also does set_live(false).
     void set_bucket(Bucket* b) { bucket_rep = reinterpret_cast<uintptr_t>(b); }
     size_t  bytes;   // Number of bytes in this allocation
+
     // Access to the allocation liveness flag (for leak checking)
-    bool live() const { return bucket_rep & 1; }
-    void set_live(bool l) { bucket_rep = (bucket_rep & ~uintptr_t(1)) | l; }
+    bool live() const { return bucket_rep & kLive; }
+    void set_live(bool l) {
+      bucket_rep = (bucket_rep & ~uintptr_t(kLive)) | (l ? kLive : 0);
+    }
+
+    // Should this allocation be ignored if it looks like a leak?
+    bool ignore() const { return bucket_rep & kIgnore; }
+    void set_ignore(bool r) {
+      bucket_rep = (bucket_rep & ~uintptr_t(kIgnore)) | (r ? kIgnore : 0);
+    }
+
    private:
-    uintptr_t bucket_rep;  // Bucket* with the lower bit being the "live" flag:
-                           // pointers point to at least 4-byte aligned storage.
+    // We store a few bits in the bottom bits of bucket_rep.
+    // (Alignment is at least four, so we have at least two bits.)
+    static const int kLive = 1;
+    static const int kIgnore = 2;
+    static const int kMask = kLive | kIgnore;
+
+    uintptr_t bucket_rep;
   };
 
   // helper for FindInsideAlloc
@@ -193,12 +233,11 @@ class HeapProfileTable {
 
   // Arguments that need to be passed DumpNonLiveIterator callback below.
   struct DumpArgs {
-    int fd;  // file to write to
-    bool dump_alloc_addresses;  // if we are dumping allocation's addresses
-    Stats* profile_stats;  // stats to update
+    RawFD fd;  // file to write to
+    Stats* profile_stats;  // stats to update (may be NULL)
 
-    DumpArgs(int a, bool b, Stats* d)
-      : fd(a), dump_alloc_addresses(b), profile_stats(d) { }
+    DumpArgs(RawFD a, Stats* d)
+      : fd(a), profile_stats(d) { }
   };
 
   // helpers ----------------------------
@@ -207,9 +246,16 @@ class HeapProfileTable {
   // We return the amount of space in buf that we use.  We start printing
   // at buf + buflen, and promise not to go beyond buf + bufsize.
   // We do not provision for 0-terminating 'buf'.
-  // We update *profile_stats by counting bucket b.
+  //
+  // If profile_stats is non-NULL, we update *profile_stats by
+  // counting bucket b.
+  //
+  // "extra" is appended to the unparsed bucket.  Typically it is empty,
+  // but may be set to something like " heapprofile" for the total
+  // bucket to indicate the type of the profile.
   static int UnparseBucket(const Bucket& b,
                            char* buf, int buflen, int bufsize,
+                           const char* extra,
                            Stats* profile_stats);
 
   // Get the bucket for the caller stack trace 'key' of depth 'depth'
@@ -224,6 +270,8 @@ class HeapProfileTable {
     info.object_size = v->bytes;
     info.call_stack = v->bucket()->stack;
     info.stack_depth = v->bucket()->depth;
+    info.live = v->live();
+    info.ignored = v->ignore();
     callback(ptr, info);
   }
 
@@ -232,10 +280,33 @@ class HeapProfileTable {
   inline static void DumpNonLiveIterator(const void* ptr, AllocValue* v,
                                          const DumpArgs& args);
 
-  // data ----------------------------
+  // Helper for IterateOrderedAllocContexts and FillOrderedProfile.
+  // Creates a sorted list of Buckets whose length is num_buckets_.
+  // The caller is responsible for dellocating the returned list.
+  Bucket** MakeSortedBucketList() const;
 
-  // Size for table_.
-  static const int kHashTableSize = 179999;
+  // Helper for TakeSnapshot.  Saves object to snapshot.
+  static void AddToSnapshot(const void* ptr, AllocValue* v, Snapshot* s);
+
+  // Arguments passed to AddIfNonLive
+  struct AddNonLiveArgs {
+    Snapshot* dest;
+    Snapshot* base;
+  };
+
+  // Helper for NonLiveSnapshot.  Adds the object to the destination
+  // snapshot if it is non-live.
+  static void AddIfNonLive(const void* ptr, AllocValue* v,
+                           AddNonLiveArgs* arg);
+
+  // Write contents of "*allocations" as a heap profile to
+  // "file_name".  "total" must contain the total of all entries in
+  // "*allocations".
+  static bool WriteProfile(const char* file_name,
+                           const Bucket& total,
+                           AllocationMap* allocations);
+
+  // data ----------------------------
 
   // Memory (de)allocator that we use.
   Allocator alloc_;
@@ -258,4 +329,56 @@ class HeapProfileTable {
   DISALLOW_EVIL_CONSTRUCTORS(HeapProfileTable);
 };
 
+class HeapProfileTable::Snapshot {
+ public:
+  const Stats& total() const { return total_; }
+
+  // Report anything in this snapshot as a leak.
+  // May use new/delete for temporary storage.
+  // If should_symbolize is true, will fork (which is not threadsafe)
+  // to turn addresses into symbol names.  Set to false for maximum safety.
+  // Also writes a heap profile to "filename" that contains
+  // all of the objects in this snapshot.
+  void ReportLeaks(const char* checker_name, const char* filename,
+                   bool should_symbolize);
+
+  // Report the addresses of all leaked objects.
+  // May use new/delete for temporary storage.
+  void ReportIndividualObjects();
+
+  bool Empty() const {
+    return (total_.allocs == 0) && (total_.alloc_size == 0);
+  }
+
+ private:
+  friend class HeapProfileTable;
+
+  // Total count/size are stored in a Bucket so we can reuse UnparseBucket
+  Bucket total_;
+
+  // We share the Buckets managed by the parent table, but have our
+  // own object->bucket map.
+  AllocationMap map_;
+
+  Snapshot(Allocator alloc, DeAllocator dealloc) : map_(alloc, dealloc) {
+    memset(&total_, 0, sizeof(total_));
+  }
+
+  // Callback used to populate a Snapshot object with entries found
+  // in another allocation map.
+  inline void Add(const void* ptr, const AllocValue& v) {
+    map_.Insert(ptr, v);
+    total_.allocs++;
+    total_.alloc_size += v.bytes;
+  }
+
+  // Helpers for sorting and generating leak reports
+  struct Entry;
+  struct ReportState;
+  static void ReportCallback(const void* ptr, AllocValue* v, ReportState*);
+  static void ReportObject(const void* ptr, AllocValue* v, char*);
+
+  DISALLOW_COPY_AND_ASSIGN(Snapshot);
+};
+
 #endif  // BASE_HEAP_PROFILE_TABLE_H_
