@@ -32,17 +32,25 @@
 //
 // TODO: Log large allocations
 
-#include "config.h"
+#include <config.h>
 
+#include <stdio.h>
 #include <stdlib.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef HAVE_INTTYPES_H
+#include <inttypes.h>
+#endif
+#ifdef HAVE_FCNTL_H
+#include <fcntl.h>    // for open()
+#endif
 #ifdef HAVE_MMAP
 #include <sys/mman.h>
 #endif
 #include <errno.h>
 #include <assert.h>
+#include <sys/types.h>
 
 #include <algorithm>
 #include <string>
@@ -54,6 +62,7 @@
 #include "base/googleinit.h"
 #include "base/commandlineflags.h"
 #include "malloc_hook-inl.h"
+#include "tcmalloc_guard.h"
 #include <google/malloc_hook.h>
 #include <google/malloc_extension.h>
 #include "base/spinlock.h"
@@ -83,13 +92,21 @@ using STL_NAMESPACE::sort;
 
 DEFINE_int64(heap_profile_allocation_interval,
              EnvToInt64("HEAP_PROFILE_ALLOCATION_INTERVAL", 1 << 30 /*1GB*/),
-             "Dump heap profiling information once every specified "
-             "number of bytes allocated by the program.");
+             "If non-zero, dump heap profiling information once every "
+             "specified number of bytes allocated by the program since "
+             "the last dump.");
+DEFINE_int64(heap_profile_deallocation_interval,
+             EnvToInt64("HEAP_PROFILE_DEALLOCATION_INTERVAL", 0),
+             "If non-zero, dump heap profiling information once every "
+             "specified number of bytes deallocated by the program "
+             "since the last dump.");
+// We could also add flags that report whenever inuse_bytes changes by
+// X or -X, but there hasn't been a need for that yet, so we haven't.
 DEFINE_int64(heap_profile_inuse_interval,
              EnvToInt64("HEAP_PROFILE_INUSE_INTERVAL", 100 << 20 /*100MB*/),
-              "Dump heap profiling information whenever the high-water "
-             "memory usage mark increases by the specified number of "
-             "bytes.");
+             "If non-zero, dump heap profiling information whenever "
+             "the high-water memory usage mark increases by the specified "
+             "number of bytes.");
 DEFINE_bool(mmap_log,
             EnvToBool("HEAP_PROFILE_MMAP_LOG", false),
             "Should mmap/munmap calls be logged?");
@@ -127,16 +144,29 @@ static void ProfilerFree(void* p) {
   LowLevelAlloc::Free(p);
 }
 
+// We use buffers of this size in DoGetHeapProfile.
+static const int kProfileBufferSize = 1 << 20;
+
+// This is a last-ditch buffer we use in DumpProfileLocked in case we
+// can't allocate more memory from ProfilerMalloc.  We expect this
+// will be used by HeapProfileEndWriter when the application has to
+// exit due to out-of-memory.  This buffer is allocated in
+// HeapProfilerStart.  Access to this must be protected by heap_lock.
+static char* global_profiler_buffer = NULL;
+
+
 //----------------------------------------------------------------------
 // Profiling control/state data
 //----------------------------------------------------------------------
 
+// Access to all of these is protected by heap_lock.
 static bool  is_on = false;           // If are on as a subsytem.
 static bool  dumping = false;         // Dumping status to prevent recursion
 static char* filename_prefix = NULL;  // Prefix used for profile file names
                                       // (NULL if no need for dumping yet)
 static int   dump_count = 0;          // How many dumps so far
-static int64 last_dump = 0;           // When did we last dump
+static int64 last_dump_alloc = 0;     // alloc_size when did we last dump
+static int64 last_dump_free = 0;      // free_size when did we last dump
 static int64 high_water_mark = 0;     // In-use-bytes at last high-water dump
 
 static HeapProfileTable* heap_profile = NULL;  // the heap profile table
@@ -169,21 +199,19 @@ static void AddRemoveMMapDataLocked(AddOrRemove mode) {
   }
 }
 
-static char* DoGetHeapProfile(void* (*alloc_func)(size_t)) {
+// Input must be a buffer of size at least 1MB.
+static char* DoGetHeapProfileLocked(char* buf, int buflen) {
   // We used to be smarter about estimating the required memory and
   // then capping it to 1MB and generating the profile into that.
-  // However it should not cost us much to allocate 1MB every time.
-  static const int size = 1 << 20;
-  char* buf = reinterpret_cast<char*>((*alloc_func)(size));
-  if (buf == NULL) return NULL;
+  if (buf == NULL || buflen < 1)
+    return NULL;
 
-  // Grab the lock and generate the profile.
-  SpinLockHolder l(&heap_lock);
-  int buflen = 0;
+  RAW_DCHECK(heap_lock.IsHeld(), "");
+  int bytes_written = 0;
   if (is_on) {
     HeapProfileTable::Stats const stats = heap_profile->total();
     AddRemoveMMapDataLocked(ADD);
-    buflen = heap_profile->FillOrderedProfile(buf, size - 1);
+    bytes_written = heap_profile->FillOrderedProfile(buf, buflen - 1);
     // FillOrderedProfile should not reduce the set of active mmap-ed regions,
     // hence MemoryRegionMap will let us remove everything we've added above:
     AddRemoveMMapDataLocked(REMOVE);
@@ -191,15 +219,17 @@ static char* DoGetHeapProfile(void* (*alloc_func)(size_t)) {
     // if this fails, we somehow removed by AddRemoveMMapDataLocked
     // more than we have added.
   }
-  buf[buflen] = '\0';
-  RAW_DCHECK(buflen == strlen(buf), "");
+  buf[bytes_written] = '\0';
+  RAW_DCHECK(bytes_written == strlen(buf), "");
 
   return buf;
 }
 
 extern "C" char* GetHeapProfile() {
   // Use normal malloc: we return the profile to the user to free it:
-  return DoGetHeapProfile(&malloc);
+  SpinLockHolder l(&heap_lock);
+  return DoGetHeapProfileLocked(
+      reinterpret_cast<char*>(malloc(kProfileBufferSize)), kProfileBufferSize);
 }
 
 // defined below
@@ -231,24 +261,28 @@ static void DumpProfileLocked(const char* reason) {
   snprintf(file_name, sizeof(file_name), "%s.%04d%s",
            filename_prefix, dump_count, HeapProfileTable::kFileExt);
 
-  // Release allocation lock around the meat of this routine
-  // thus not blocking other threads too much.
-  heap_lock.Unlock();
-  {
-    // Dump the profile
-    RAW_VLOG(0, "Dumping heap profile to %s (%s)", file_name, reason);
-    FILE* f = fopen(file_name, "w");
-    if (f != NULL) {
-      // Use ProfilerMalloc not to put info about profile itself into itself:
-      char* profile = DoGetHeapProfile(&ProfilerMalloc);
-      fputs(profile, f);
-      ProfilerFree(profile);
-      fclose(f);
-    } else {
-      RAW_LOG(ERROR, "Failed dumping heap profile to %s", file_name);
-    }
+  // Dump the profile
+  RAW_VLOG(0, "Dumping heap profile to %s (%s)", file_name, reason);
+  // We must use file routines that don't access memory, since we hold
+  // a memory lock now.
+  RawFD fd = RawOpenForWriting(file_name);
+  if (fd == kIllegalRawFD) {
+    RAW_LOG(ERROR, "Failed dumping heap profile to %s", file_name);
+    dumping = false;
+    return;
   }
-  heap_lock.Lock();
+
+  // This case may be impossible, but it's best to be safe.
+  // It's safe to use the global buffer: we're protected by heap_lock.
+  if (global_profiler_buffer == NULL) {
+    global_profiler_buffer =
+        reinterpret_cast<char*>(ProfilerMalloc(kProfileBufferSize));
+  }
+
+  char* profile = DoGetHeapProfileLocked(global_profiler_buffer,
+                                         kProfileBufferSize);
+  RawWrite(fd, profile, strlen(profile));
+  RawClose(fd);
 
   dumping = false;
 }
@@ -257,34 +291,52 @@ static void DumpProfileLocked(const char* reason) {
 // Profile collection
 //----------------------------------------------------------------------
 
+// Dump a profile after either an allocation or deallocation, if
+// the memory use has changed enough since the last dump.
+static void MaybeDumpProfileLocked() {
+  if (!dumping) {
+    const HeapProfileTable::Stats& total = heap_profile->total();
+    const int64 inuse_bytes = total.alloc_size - total.free_size;
+    bool need_to_dump = false;
+    char buf[128];
+    if (FLAGS_heap_profile_allocation_interval > 0 &&
+        total.alloc_size >=
+        last_dump_alloc + FLAGS_heap_profile_allocation_interval) {
+      snprintf(buf, sizeof(buf), ("%"PRId64" MB allocated cumulatively, "
+                                  "%"PRId64" MB currently in use"),
+               total.alloc_size >> 20, inuse_bytes >> 20);
+      need_to_dump = true;
+    } else if (FLAGS_heap_profile_deallocation_interval > 0 &&
+               total.free_size >=
+               last_dump_free + FLAGS_heap_profile_deallocation_interval) {
+      snprintf(buf, sizeof(buf), ("%"PRId64" MB freed cumulatively, "
+                                  "%"PRId64" MB currently in use"),
+               total.free_size >> 20, inuse_bytes >> 20);
+      need_to_dump = true;
+    } else if (FLAGS_heap_profile_inuse_interval > 0 &&
+               inuse_bytes >
+               high_water_mark + FLAGS_heap_profile_inuse_interval) {
+      snprintf(buf, sizeof(buf), "%"PRId64" MB currently in use",
+               inuse_bytes >> 20);
+      need_to_dump = true;
+    }
+    if (need_to_dump) {
+      DumpProfileLocked(buf);
+
+      last_dump_alloc = total.alloc_size;
+      last_dump_free = total.free_size;
+      if (inuse_bytes > high_water_mark)
+        high_water_mark = inuse_bytes;
+    }
+  }
+}
+
 // Record an allocation in the profile.
 static void RecordAlloc(const void* ptr, size_t bytes, int skip_count) {
   SpinLockHolder l(&heap_lock);
   if (is_on) {
     heap_profile->RecordAlloc(ptr, bytes, skip_count + 1);
-    const HeapProfileTable::Stats& total = heap_profile->total();
-    const int64 inuse_bytes = total.alloc_size - total.free_size;
-    if (!dumping) {
-      bool need_to_dump = false;
-      char buf[128];
-      if (total.alloc_size >=
-          last_dump + FLAGS_heap_profile_allocation_interval) {
-        snprintf(buf, sizeof(buf), "%"PRId64" MB allocated",
-                 total.alloc_size >> 20);
-        // Track that we made a "total allocation size" dump
-        last_dump = total.alloc_size;
-        need_to_dump = true;
-      } else if (inuse_bytes >
-                 high_water_mark + FLAGS_heap_profile_inuse_interval) {
-        snprintf(buf, sizeof(buf), "%"PRId64" MB in use", inuse_bytes >> 20);
-        // Track that we made a "high water mark" dump
-        high_water_mark = inuse_bytes;
-        need_to_dump = true;
-      }
-      if (need_to_dump) {
-        DumpProfileLocked(buf);
-      }
-    }
+    MaybeDumpProfileLocked();
   }
 }
 
@@ -293,6 +345,7 @@ static void RecordFree(const void* ptr) {
   SpinLockHolder l(&heap_lock);
   if (is_on) {
     heap_profile->RecordFree(ptr);
+    MaybeDumpProfileLocked();
   }
 }
 
@@ -331,8 +384,8 @@ static void MmapHook(const void* result, const void* start, size_t size,
     // in pretty-printing of NULL as "nil".
     // TODO(maxim): instead should use a safe snprintf reimplementation
     RAW_LOG(INFO,
-            "mmap(start=0x%"PRIxS", len=%"PRIuS", prot=0x%x, flags=0x%x, "
-            "fd=%d, offset=0x%x) = 0x%"PRIxS"",
+            "mmap(start=0x%"PRIxPTR", len=%"PRIuS", prot=0x%x, flags=0x%x, "
+            "fd=%d, offset=0x%x) = 0x%"PRIxPTR"",
             (uintptr_t) start, size, prot, flags, fd, (unsigned int) offset,
             (uintptr_t) result);
 #ifdef TODO_REENABLE_STACK_TRACING
@@ -354,8 +407,9 @@ static void MremapHook(const void* result, const void* old_addr,
     // in pretty-printing of NULL as "nil".
     // TODO(maxim): instead should use a safe snprintf reimplementation
     RAW_LOG(INFO,
-            "mremap(old_addr=0x%"PRIxS", old_size=%"PRIuS", new_size=%"PRIuS", "
-            "flags=0x%x, new_addr=0x%"PRIxS") = 0x%"PRIxS"",
+            "mremap(old_addr=0x%"PRIxPTR", old_size=%"PRIuS", "
+            "new_size=%"PRIuS", flags=0x%x, new_addr=0x%"PRIxPTR") = "
+            "0x%"PRIxPTR"",
             (uintptr_t) old_addr, old_size, new_size, flags,
             (uintptr_t) new_addr, (uintptr_t) result);
 #ifdef TODO_REENABLE_STACK_TRACING
@@ -372,7 +426,7 @@ static void MunmapHook(const void* ptr, size_t size) {
     // We use PRIxS not just '%p' to avoid deadlocks
     // in pretty-printing of NULL as "nil".
     // TODO(maxim): instead should use a safe snprintf reimplementation
-    RAW_LOG(INFO, "munmap(start=0x%"PRIxS", len=%"PRIuS")",
+    RAW_LOG(INFO, "munmap(start=0x%"PRIxPTR", len=%"PRIuS")",
                   (uintptr_t) ptr, size);
 #ifdef TODO_REENABLE_STACK_TRACING
     DumpStackTrace(1, RawInfoStackDumper, NULL);
@@ -385,7 +439,7 @@ static void MunmapHook(const void* ptr, size_t size) {
 
 static void SbrkHook(const void* result, ptrdiff_t increment) {
   if (FLAGS_mmap_log) {  // log it
-    RAW_LOG(INFO, "sbrk(inc=%"PRIdS") = 0x%"PRIxS"",
+    RAW_LOG(INFO, "sbrk(inc=%"PRIdS") = 0x%"PRIxPTR"",
                   increment, (uintptr_t) result);
 #ifdef TODO_REENABLE_STACK_TRACING
     DumpStackTrace(1, RawInfoStackDumper, NULL);
@@ -409,6 +463,10 @@ extern "C" void HeapProfilerStart(const char* prefix) {
 
   RAW_VLOG(0, "Starting tracking the heap");
 
+  // This should be done before the hooks are set up, since it should
+  // call new, and we want that to be accounted for correctly.
+  MallocExtension::Initialize();
+
   if (FLAGS_only_mmap_profile) {
     FLAGS_mmap_profile = true;
   }
@@ -431,10 +489,17 @@ extern "C" void HeapProfilerStart(const char* prefix) {
   heap_profiler_memory =
     LowLevelAlloc::NewArena(0, LowLevelAlloc::DefaultArena());
 
+  // Reserve space now for the heap profiler, so we can still write a
+  // heap profile even if the application runs out of memory.
+  global_profiler_buffer =
+      reinterpret_cast<char*>(ProfilerMalloc(kProfileBufferSize));
+
   heap_profile = new(ProfilerMalloc(sizeof(HeapProfileTable)))
                    HeapProfileTable(ProfilerMalloc, ProfilerFree);
 
-  last_dump = 0;
+  last_dump_alloc = 0;
+  last_dump_free = 0;
+  high_water_mark = 0;
 
   // We do not reset dump_count so if the user does a sequence of
   // HeapProfilerStart/HeapProfileStop, we will get a continuous
@@ -457,10 +522,11 @@ extern "C" void HeapProfilerStart(const char* prefix) {
   filename_prefix = reinterpret_cast<char*>(ProfilerMalloc(prefix_length + 1));
   memcpy(filename_prefix, prefix, prefix_length);
   filename_prefix[prefix_length] = '\0';
+}
 
-  // This should be done before the hooks are set up, since it should
-  // call new, and we want that to be accounted for correctly.
-  MallocExtension::Initialize();
+extern "C" bool IsHeapProfilerRunning() {
+  SpinLockHolder l(&heap_lock);
+  return is_on;
 }
 
 extern "C" void HeapProfilerStop() {
@@ -494,6 +560,9 @@ extern "C" void HeapProfilerStop() {
   ProfilerFree(heap_profile);
   heap_profile = NULL;
 
+  // free output-buffer memory
+  ProfilerFree(global_profiler_buffer);
+
   // free prefix
   ProfilerFree(filename_prefix);
   filename_prefix = NULL;
@@ -546,5 +615,7 @@ struct HeapProfileEndWriter {
   ~HeapProfileEndWriter() { HeapProfilerDump("Exiting"); }
 };
 
+// We want to make sure tcmalloc is up and running before starting the profiler
+static const TCMallocGuard tcmalloc_initializer;
 REGISTER_MODULE_INITIALIZER(heapprofiler, HeapProfilerInit());
 static HeapProfileEndWriter heap_profile_end_writer;
