@@ -79,8 +79,6 @@
 #include <malloc.h>
 #endif
 
-#include <netinet/in.h>         // inet_ntoa
-#include <arpa/inet.h>          // inet_ntoa
 #ifdef HAVE_EXECINFO_H
 #include <execinfo.h>           // backtrace
 #endif
@@ -103,6 +101,7 @@
 #include "base/googleinit.h"
 #include "base/logging.h"
 #include "base/commandlineflags.h"
+#include "base/thread_lister.h"
 #include <google/heap-checker.h>
 #include "memory_region_map.h"
 #include <google/malloc_extension.h>
@@ -257,6 +256,9 @@ void* operator new[](size_t size, const Initialized&) {
   return p;
 }
 
+static void DoWipeStack(int n);  // defined below
+static void WipeStack() { DoWipeStack(20); }
+
 static void Pause() {
   poll(NULL, 0, 77);  // time for thread activity in HeapBusyThreadBody
 
@@ -275,6 +277,8 @@ static void Pause() {
       }
     }
   }
+  WipeStack();  // e.g. MallocExtension::VerifyAllMemory
+                // can leave pointers to heap objects on stack
 }
 
 // Make gcc think a pointer is "used"
@@ -315,30 +319,28 @@ static void LogHidden(const char* message, const void* ptr) {
        << ptr << " ^ " << reinterpret_cast<void*>(kHideMask) << endl;
 }
 
-// non-static to fool the compiler against inlining
-extern void (*run_hidden_ptr)(Closure* c, int n);
-void (*run_hidden_ptr)(Closure* c, int n);
-extern void (*wipe_stack_ptr)(int n);
-void (*wipe_stack_ptr)(int n);
+// volatile to fool the compiler against inlining the calls to these
+void (*volatile run_hidden_ptr)(Closure* c, int n);
+void (*volatile wipe_stack_ptr)(int n);
 
 static void DoRunHidden(Closure* c, int n) {
   if (n) {
     VLOG(10) << "Level " << n << " at " << &n;
-    run_hidden_ptr(c, n-1);
-    wipe_stack_ptr(n);
+    (*run_hidden_ptr)(c, n-1);
+    (*wipe_stack_ptr)(n);
     sleep(0);  // undo -foptimize-sibling-calls
   } else {
     c->Run();
   }
 }
 
-static void DoWipeStack(int n) {
+/*static*/ void DoWipeStack(int n) {
   VLOG(10) << "Wipe level " << n << " at " << &n;
   if (n) {
     const int sz = 30;
     volatile int arr[sz];
-    for (int i = 0; i < sz; ++i)  arr[i] = 0;
-    wipe_stack_ptr(n-1);
+    for (int i = 0; i < sz; ++i) arr[i] = 0;
+    (*wipe_stack_ptr)(n-1);
     sleep(0);  // undo -foptimize-sibling-calls
   }
 }
@@ -402,6 +404,7 @@ enum CheckType { SAME_HEAP, NO_LEAKS };
 
 static void VerifyLeaks(HeapLeakChecker* check, CheckType type,
                         int leaked_bytes, int leaked_objects) {
+  WipeStack();  // to help with can_create_leaks_reliably
   const bool no_leaks =
     type == NO_LEAKS ? RUN_SILENT(*check, BriefNoLeaks)
                      : RUN_SILENT(*check, BriefSameHeap);
@@ -474,7 +477,9 @@ static void TestHeapLeakCheckerDeathInverse() {
   LogHidden("Leaking", foo);
   DeAllocHidden(&bar);
   Pause();
-  VerifyLeaks(&check, SAME_HEAP, -150 * static_cast<int64>(sizeof(int)), 0);
+  VerifyLeaks(&check, SAME_HEAP,
+              100 * static_cast<int64>(sizeof(int)),
+              1);
   DeAllocHidden(&foo);
 }
 
@@ -506,7 +511,9 @@ static void TestHeapLeakCheckerDeathCountLess() {
   DeAllocHidden(&bar1);
   DeAllocHidden(&bar2);
   Pause();
-  VerifyLeaks(&check, SAME_HEAP, 0, -1);
+  VerifyLeaks(&check, SAME_HEAP,
+              100 * sizeof(int),
+              1);
   DeAllocHidden(&foo);
 }
 
@@ -525,7 +532,9 @@ static void TestHeapLeakCheckerDeathCountMore() {
   LogHidden("Leaking", bar2);
   DeAllocHidden(&foo);
   Pause();
-  VerifyLeaks(&check, SAME_HEAP, 0, 1);
+  VerifyLeaks(&check, SAME_HEAP,
+              100 * sizeof(int),
+              2);
   DeAllocHidden(&bar1);
   DeAllocHidden(&bar2);
 }
@@ -552,8 +561,8 @@ static void TestHeapLeakChecker() {
   }
 }
 
-// no false positives from pprof
-static void TestHeapLeakCheckerPProf() {
+// no false positives
+static void TestHeapLeakCheckerNoFalsePositives() {
   { HeapLeakChecker check("trivial_p");
     int foo = 5;
     int* p = &foo;
@@ -574,9 +583,9 @@ static void TestHeapLeakCheckerPProf() {
   }
 }
 
-// trick heap change: same total # of bytes and objects, but
-// different individual object sizes
-static void TestHeapLeakCheckerTrick() {
+// test that we detect leaks when we have same total # of bytes and
+// objects, but different individual object sizes
+static void TestLeakButTotalsMatch() {
   void* bar1 = AllocHidden(240 * sizeof(int));
   Use(&bar1);
   void* bar2 = AllocHidden(160 * sizeof(int));
@@ -594,15 +603,10 @@ static void TestHeapLeakCheckerTrick() {
   DeAllocHidden(&bar1);
   DeAllocHidden(&bar2);
   Pause();
-  if (can_create_leaks_reliably) {
-    // this might still fail occasionally, but it should be very rare:
-    CHECK(check.BriefSameHeap());
-  } else {
-    // we expect it to usually misbehave, so we silence it:
-    WARN_IF(RUN_SILENT(check, BriefSameHeap) != true,
-            "Tricky leaks unexpectedly found: "
-            "Some liveness flood must be too optimistic");
-  }
+
+  // foo1 and foo2 leaked
+  VerifyLeaks(&check, NO_LEAKS, (280+120)*sizeof(int), 2);
+
   DeAllocHidden(&foo1);
   DeAllocHidden(&foo2);
 }
@@ -638,32 +642,17 @@ static void TransLeaks() {
   AllocHidden(1 * sizeof(char));
 }
 
-// have leaks but disable them
-static void DisabledLeaks() {
-  HeapLeakChecker::DisableChecksUp(1);
+// range-based disabling using Disabler
+static void ScopedDisabledLeaks() {
+  HeapLeakChecker::Disabler disabler;
   AllocHidden(3 * sizeof(int));
   TransLeaks();
+  malloc(10);  // Direct leak
 }
 
-// have leaks but range-disable them
-static void RangeDisabledLeaks() {
-  void* start_address = HeapLeakChecker::GetDisableChecksStart();
-  AllocHidden(3 * sizeof(int));
-  TransLeaks();
-  HeapLeakChecker::DisableChecksToHereFrom(start_address);
-}
-
-// We need this function pointer trickery to fool an aggressive
-// optimizing compiler such as icc into not inlining DisabledLeaks().
-// Otherwise the stack-frame-address-based disabling in it
-// will wrongly disable allocation tracking in
-// the functions into which it's inlined.
-static void (*disabled_leaks_addr)() = &DisabledLeaks;
-
 // have different disabled leaks
 static void* RunDisabledLeaks(void* a) {
-  disabled_leaks_addr();
-  RangeDisabledLeaks();
+  ScopedDisabledLeaks();
   return a;
 }
 
@@ -689,10 +678,6 @@ static void TestHeapLeakCheckerDisabling() {
   ThreadDisabledLeaks();
   ThreadDisabledLeaks();
 
-  LOG_IF(FATAL, HeapLeakChecker::HaveDisabledChecksUp(1),
-        "Some code with DisableChecksUp() got inlined into here; "
-        "need to add more tricks to prevent this inlining.");
-
   Pause();
 
   CHECK(check.SameHeap());
@@ -747,6 +732,7 @@ static void TestSTLAllocInverse() {
   RunHidden(NewCallback(DoTestSTLAllocInverse, &x));
   LogHidden("Leaking", x);
   if (can_create_leaks_reliably) {
+    WipeStack();  // to help with can_create_leaks_reliably
     // these might still fail occasionally, but it should be very rare
     CHECK_EQ(RUN_SILENT(check, BriefNoLeaks), false);
     CHECK_GE(check.BytesLeaked(), 100 * sizeof(int));
@@ -818,9 +804,6 @@ static void TestLibCAllocate() {
   }
 
   strerror(errno);
-  struct in_addr addr;
-  addr.s_addr = INADDR_ANY;
-  inet_ntoa(addr);
   const time_t now = time(NULL);
   ctime(&now);
 #ifdef HAVE_EXECINFO_H
@@ -904,7 +887,7 @@ static void* HeapBusyThreadBody(void* a) {
       ptr = reinterpret_cast<int **>(
           reinterpret_cast<uintptr_t>(ptr) ^ kHideMask);
       // busy loop to get the thread interrupted at:
-      for (int i = 1; i < 1000000; ++i)  user += (1 + user * user * 5) / i;
+      for (int i = 1; i < 10000000; ++i)  user += (1 + user * user * 5) / i;
       ptr = reinterpret_cast<int **>(
           reinterpret_cast<uintptr_t>(ptr) ^ kHideMask);
     } else {
@@ -941,114 +924,6 @@ static void RunHeapBusyThreads() {
   Pause();
 }
 
-// tests disabling via function name
-REGISTER_MODULE_INITIALIZER(heap_checker_unittest, {
-  // TODO(csilvers): figure out when this might be true
-  can_create_leaks_reliably = false;
-  HeapLeakChecker::DisableChecksIn("NamedDisabledLeaks");
-});
-
-// NOTE: For NamedDisabledLeaks, NamedTwoDisabledLeaks
-// and NamedThreeDisabledLeaks for the name-based disabling to work in opt mode
-// we need to undo the tail-recursion optimization effect
-// of -foptimize-sibling-calls that is enabled by -O2 in gcc 3.* and 4.*
-// so that for the leaking calls we can find the stack frame
-// that resolves to a Named*DisabledLeaks function.
-// We do this by adding a fake last statement to these functions
-// so that tail-recursion optimization is done with it.
-
-// have leaks that we disable via our function name in MODULE_INITIALIZER
-static void NamedDisabledLeaks() {
-  AllocHidden(5 * sizeof(float));
-  TransLeaks();
-  sleep(0);  // undo -foptimize-sibling-calls
-}
-
-// to trick complier into preventing inlining
-static void (*named_disabled_leaks)() = &NamedDisabledLeaks;
-
-// have leaks that we disable via our function name ourselves
-static void NamedTwoDisabledLeaks() {
-  static bool first = true;
-  if (first) {
-    HeapLeakChecker::DisableChecksIn("NamedTwoDisabledLeaks");
-    first = false;
-  }
-  AllocHidden(5 * sizeof(double));
-  TransLeaks();
-  sleep(0);  // undo -foptimize-sibling-calls
-}
-
-// to trick complier into preventing inlining
-static void (*named_two_disabled_leaks)() = &NamedTwoDisabledLeaks;
-
-// have leaks that we disable via our function name in our caller
-static void NamedThreeDisabledLeaks() {
-  AllocHidden(5 * sizeof(float));
-  TransLeaks();
-  sleep(0);  // undo -foptimize-sibling-calls
-}
-
-// to trick complier into preventing inlining
-static void (*named_three_disabled_leaks)() = &NamedThreeDisabledLeaks;
-
-static bool range_disable_named = false;
-
-// have leaks that we disable via function names
-static void* RunNamedDisabledLeaks(void* a) {
-  // We get the address unconditionally here to fool gcc 4.1.0 in opt mode:
-  // else it reorders the binary code so that our return address bracketing
-  // does not work here.
-  void* start_address = HeapLeakChecker::GetDisableChecksStart();
-
-  named_disabled_leaks();
-  named_two_disabled_leaks();
-  named_three_disabled_leaks();
-
-  // TODO(maxim): do not need this if we make pprof work in automated test runs
-  if (range_disable_named) {
-    HeapLeakChecker::DisableChecksToHereFrom(start_address);
-  }
-  sleep(0);  // undo -foptimize-sibling-calls
-  return a;
-}
-
-// have leaks inside of threads that we disable via function names
-static void ThreadNamedDisabledLeaks() {
-  if (FLAGS_no_threads)  return;
-  pthread_t tid;
-  pthread_attr_t attr;
-  CHECK_EQ(pthread_attr_init(&attr), 0);
-  CHECK_EQ(pthread_create(&tid, &attr, RunNamedDisabledLeaks, NULL), 0);
-  void* res;
-  CHECK_EQ(pthread_join(tid, &res), 0);
-}
-
-// test leak disabling via function names
-static void TestHeapLeakCheckerNamedDisabling() {
-  HeapLeakChecker::DisableChecksIn("NamedThreeDisabledLeaks");
-
-  HeapLeakChecker check("named_disabling");
-
-  RunNamedDisabledLeaks(NULL);
-  RunNamedDisabledLeaks(NULL);
-  ThreadNamedDisabledLeaks();
-  RunNamedDisabledLeaks(NULL);
-  ThreadNamedDisabledLeaks();
-  ThreadNamedDisabledLeaks();
-
-  Pause();
-
-  if (!FLAGS_maybe_stripped) {
-    CHECK_EQ(check.SameHeap(), true);
-      // pprof checking should allow it
-  } else {
-    WARN_IF(check.SameHeap() != true,
-            "named_disabling leaks are caught; "
-            "we must be using a stripped binary");
-  }
-}
-
 // ========================================================================= //
 
 // This code section is to test that objects that are reachable from global
@@ -1490,7 +1365,10 @@ int main(int argc, char** argv) {
   if (FLAGS_test_register_leak) {
     // make us fail only where the .sh test expects:
     Pause();
-    for (int i = 0; i < 20; ++i) CHECK(HeapLeakChecker::NoGlobalLeaks());
+    for (int i = 0; i < 100; ++i) {  // give it some time to crash
+      CHECK(HeapLeakChecker::NoGlobalLeaks());
+      Pause();
+    }
     return Pass();
   }
 
@@ -1500,7 +1378,7 @@ int main(int argc, char** argv) {
 
   TestHeapLeakChecker();
   Pause();
-  TestHeapLeakCheckerTrick();
+  TestLeakButTotalsMatch();
   Pause();
 
   TestHeapLeakCheckerDeathSimple();
@@ -1521,7 +1399,7 @@ int main(int argc, char** argv) {
 
   CHECK(HeapLeakChecker::NoGlobalLeaks());  // so far, so good
 
-  TestHeapLeakCheckerPProf();
+  TestHeapLeakCheckerNoFalsePositives();
   Pause();
 
   TestHeapLeakCheckerDisabling();
@@ -1555,9 +1433,6 @@ int main(int argc, char** argv) {
 
   CHECK(HeapLeakChecker::NoGlobalLeaks());  // so far, so good
 
-  void* start_address = HeapLeakChecker::GetDisableChecksStart();
-
-  TestHeapLeakCheckerNamedDisabling();
   Pause();
 
   if (!FLAGS_maybe_stripped) {
@@ -1567,14 +1442,7 @@ int main(int argc, char** argv) {
             "overall leaks are caught; we must be using a stripped binary");
   }
 
-  // TODO(maxim): do not need these if we make pprof work in automated test runs
-  HeapLeakChecker::DisableChecksToHereFrom(start_address);
-  // This will also disable (w/o relying on pprof anymore)
-  // all leaks that earlier occured inside of ThreadNamedDisabledLeaks:
-  range_disable_named = true;
-  ThreadNamedDisabledLeaks();
-
   CHECK(HeapLeakChecker::NoGlobalLeaks());  // so far, so good
-
+  
   return Pass();
 }
